package SecureDistributedKeyGeneration;

import Communication.Network;
import JointFeldmanProtocol.DistributedKeyGenerationUserImpl;
import ShamirSecretSharing.Polynomial;
import meerkat.protobuf.DKGMessages;

import java.math.BigInteger;

/**
 * Created by Tzlil on 2/22/2016.
 */
public class SecureDistributedKeyGenerationUserImpl extends DistributedKeyGenerationUserImpl {

    private final SecureDistributedKeyGeneration sdkg;

    private final Polynomial.Point[] sharesT;
    private final BigInteger[][] verificationValuesTable;

    public SecureDistributedKeyGenerationUserImpl(SecureDistributedKeyGeneration sdkg, Network network) {
        super(sdkg, network);
        this.sdkg = sdkg;
        this.sharesT = new Polynomial.Point[n];
        this.verificationValuesTable = new BigInteger[n][t + 1];
        this.messageHandler = new MessageHandler();
        this.user.setMessageHandler(this.messageHandler);
    }

    /**
     * stage1 according to the protocol
     *  1. Pi broadcasts Cik=Aik*Bik for k = 0,...,t.
     *  2. Pi computes the shares Sij,Sij' for j = 1,...,n and sends Sij,Sij' secretly to Pj.
     */
    @Override
    protected void stage1() {
        sdkg.broadcastVerificationValues(user);
        sdkg.sendSecrets(user);
    }

    /**
     * stage2 according to the protocol
     *  Pj verifies all the shares,sharesT he received
     *  if check fails for an index i, Pj broadcasts a complaint against Pi.
     *  Pj broadcasts done message at the end of this stage
     */
    @Override
    protected void stage2(){
        sdkg.setShares(shares);
        sdkg.setSharesT(sharesT);
        sdkg.broadcastComplains(user,verificationValuesTable);
        //broadcast done message after all complaints
        DKGMessages.DoneMessage doneMessage = DKGMessages.DoneMessage.newBuilder().build();
        user.broadcast(DKGMessages.Mail.Type.DONE,doneMessage);
    }

    private void ys(){
        sdkg.broadcastCommitments(user);
        //wait for receive all commitments from all i in QUAL
        for (int i:QUAL) {
            for(int k = 0; k <= t; k++) {
                while (commitmentsTable[i - 1][k] == null) {
                    try {
                        Thread.sleep(300);
                    } catch (InterruptedException e) {
                        // do nothing
                    }
                }
            }
        }
        sdkg.broadcastComplaints(user,commitmentsTable,true);
        //do something with complaints' answers

    }

    @Override
    protected void stage4() {
        ys();
        super.stage4();
    }

    private class MessageHandler extends DistributedKeyGenerationUserImpl.MessageHandler{

        final int NumberOfCommitmentsInStage1 = n * (t + 1);

        @Override
        public void handelCommitmentMessage(int sender, boolean isBroadcast, DKGMessages.CommitmentMessage commitmentMessage) {
            if(isBroadcast){
                if(commitmentsCounter < NumberOfCommitmentsInStage1){
                    int i = sender - 1;
                    int k = commitmentMessage.getK();
                    if(verificationValuesTable[i][k] == null){
                        verificationValuesTable[i][k] = extractCommitment(commitmentMessage);
                        commitmentsCounter++;
                    }
                } else{
                    super.handelCommitmentMessage(sender,isBroadcast,commitmentMessage);
                }
            }
        }

        @Override
        public void handelSecretMessage(int sender, boolean isBroadcast, DKGMessages.SecretMessage secretMessage) {
            //there is no secret message in this protocol
        }

        @Override
        public void handelDoubleSecretMessage(int sender, boolean isBroadcast, DKGMessages.DoubleSecretMessage doubleSecretMessage) {

            Polynomial.Point secret = extractSecret(doubleSecretMessage.getS1());
            Polynomial.Point secretT = extractSecret(doubleSecretMessage.getS2());
            if(!isBroadcast){
                if(shares[sender - 1] == null) {
                    shares[sender - 1] = secret;
                    sharesT[sender - 1] = secretT;
                    secretsCounter++;
                }
            }else {
                if(commitmentsCounter <= NumberOfCommitmentsInStage1) {
                    int i = sender;
                    int j = secret.x.intValue();
                    switch (complaintsTable[i - 1][j - 1]) {
                        case Waiting:
                            if (sdkg.isValidSecret(secret,secretT, verificationValuesTable[j - 1], i)) {
                                complaintsTable[i - 1][j - 1] = ComplainState.NonDisqualified;
                            } else {
                                complaintsTable[i - 1][j - 1] = ComplainState.Disqualified;
                            }
                            break;
                        default:
                            break;
                    }
                }else{ // stage4

                }
            }
        }
    }
}
